@@ -56,7 +56,9 @@ quotesyntax_obj {
 
 cpointer_obj {
  mark:
-  gcMARK(SCHEME_CPTR_VAL(p));
+  if (!(SCHEME_CPTR_FLAGS(p) & 0x1)) {
+    gcMARK(SCHEME_CPTR_VAL(p));
+  }
   gcMARK(SCHEME_CPTR_TYPE(p));
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Cptr));
@@ -64,19 +66,14 @@ cpointer_obj {
 
 offset_cpointer_obj {
  mark:
-  gcMARK(SCHEME_CPTR_VAL(p));
+  if (!(SCHEME_CPTR_FLAGS(p) & 0x1)) {
+    gcMARK(SCHEME_CPTR_VAL(p));
+  }
   gcMARK(SCHEME_CPTR_TYPE(p));
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Offset_Cptr));
 }
 
-second_of_cons {
- mark:
-  gcMARK(SCHEME_PTR2_VAL((Scheme_Object *)p));
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Simple_Object));
-}
-
 twoptr_obj {
  mark:
   gcMARK(SCHEME_PTR1_VAL((Scheme_Object *)p));
@@ -549,6 +546,15 @@ vector_obj {
 		    + ((vec->size - 1) * sizeof(Scheme_Object *))));
 }
 
+flvector_obj {
+  Scheme_Double_Vector *vec = (Scheme_Double_Vector *)p;
+
+ mark:
+ size:
+  gcBYTES_TO_WORDS((sizeof(Scheme_Double_Vector) 
+		    + ((vec->size - 1) * sizeof(double))));
+}
+
 input_port {
  mark:
   Scheme_Input_Port *ip = (Scheme_Input_Port *)p;
@@ -1188,6 +1194,16 @@ mark_sfs_info {
   gcBYTES_TO_WORDS(sizeof(SFS_Info));
 }
 
+mark_once_used {
+ mark:
+  Scheme_Once_Used *o = (Scheme_Once_Used *)p;
+  gcMARK(o->expr);
+  gcMARK(o->info);
+  gcMARK(o->next);
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Once_Used));
+}
+
 END env;
 
 /**********************************************************************/
@@ -1256,6 +1272,7 @@ mark_closure_info {
   
   gcMARK(i->local_flags);
   gcMARK(i->base_closure_map);
+  gcMARK(i->flonum_map);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Closure_Info));
@@ -1340,10 +1357,23 @@ START places;
 place_val {
  mark:
   Scheme_Place *pr = (Scheme_Place *)p;
+  gcMARK(pr->channel);
+
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Place));
 }
 
+place_async_channel_val {
+ mark:
+  Scheme_Place_Async_Channel *pac = (Scheme_Place_Async_Channel *)p;
+  int i;
+  for (i = pac->size; i--; )
+    gcMARK(pac->msgs[i]);
+
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Place_Async_Channel));
+}
+
 END places;
 
 /**********************************************************************/
@@ -1463,7 +1493,7 @@ mark_input_fd {
 }
 #endif
 
-#if defined(UNIX_PROCESSES)
+#if defined(UNIX_PROCESSES) && !(defined(MZ_USE_PLACES) && defined(MZ_PRECISE_GC))
 mark_system_child {
  mark:
   System_Child *sc = (System_Child *)p;
@@ -2221,4 +2251,51 @@ END jit;
 
 /**********************************************************************/
 
+START future;
+
+#ifdef FUTURES_ENABLED
+
+future {
+ mark:
+  future_t *f = (future_t *)p;
+  gcMARK(f->orig_lambda);
+  gcMARK(f->arg_s0);
+  gcMARK(f->arg_S0);
+  gcMARK(f->arg_b0);
+  gcMARK(f->arg_n0);
+  gcMARK(f->arg_s1);
+  gcMARK(f->arg_S1);
+  gcMARK(f->arg_s2);
+  gcMARK(f->arg_S2);
+  gcMARK(f->retval_s);
+  gcMARK(f->retval);
+  gcMARK(f->multiple_array);
+  gcMARK(f->tail_rator);
+  gcMARK(f->tail_rands);
+  gcMARK(f->prev);
+  gcMARK(f->next);
+  gcMARK(f->next_waiting_atomic);
+ size:
+  gcBYTES_TO_WORDS(sizeof(future_t));
+}
+
+#else
+
+sequential_future {
+ mark:
+  future_t *f = (future_t *)p;
+  gcMARK(f->orig_lambda);
+  gcMARK(f->running_sema);
+  gcMARK(f->retval);
+  gcMARK(f->multiple_array);
+ size:
+  gcBYTES_TO_WORDS(sizeof(future_t));
+}
+
+#endif
+
+END future;
+
+/**********************************************************************/
+
 #define GC_REG_TRAV(type, base) GC_register_traversers(type, base ## _SIZE, base ## _MARK, base ## _FIXUP, base ## _IS_CONST_SIZE, base ## _IS_ATOMIC)
