@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2009 PLT Scheme Inc.
+  Copyright (c) 2004-2010 PLT Scheme Inc.
   Copyright (c) 2000-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -59,7 +59,6 @@ static int mzerrno = 0;
 #else
 # include <errno.h>
 #endif
-#include "schfd.h"
 
 #ifdef USE_UNIX_SOCKETS_TCP
 # include <netinet/in.h>
@@ -93,6 +92,8 @@ struct SOCKADDR_IN {
 extern int scheme_stupid_windows_machine;
 #endif
 
+#include "schfd.h"
+
 #define TCP_BUFFER_SIZE 4096
 
 #ifdef USE_UNIX_SOCKETS_TCP
@@ -157,13 +158,19 @@ typedef struct Scheme_UDP {
 
 #if defined(WINDOWS_PROCESSES) || defined(WINDOWS_FILE_HANDLES)
 # define DECL_OS_FDSET(n) fd_set n[1]
-# define INIT_DECL_OS_FDSET(n) /* empty */
+# define INIT_DECL_OS_FDSET(r, w, e) /* empty */
+# define INIT_DECL_OS_RD_FDSET(r) /* empty */
+# define INIT_DECL_OS_WR_FDSET(r) /* empty */
+# define INIT_DECL_OS_ER_FDSET(r) /* empty */
 # define MZ_OS_FD_ZERO(p) FD_ZERO(p)
 # define MZ_OS_FD_SET(n, p) FD_SET(n, p)
 # define MZ_OS_FD_CLR(n, p) FD_CLR(n, p)
 #else
 # define DECL_OS_FDSET(n) DECL_FDSET(n, 1)
-# define INIT_DECL_OS_FDSET(n) INIT_DECL_FDSET(n, 1)
+# define INIT_DECL_OS_FDSET(r, w, e) INIT_DECL_FDSET(r, w, e)
+# define INIT_DECL_OS_RD_FDSET(r) INIT_DECL_RD_FDSET(r)
+# define INIT_DECL_OS_WR_FDSET(r) INIT_DECL_WR_FDSET(r)
+# define INIT_DECL_OS_ER_FDSET(r) INIT_DECL_ER_FDSET(r)
 # define MZ_OS_FD_ZERO(p) MZ_FD_ZERO(p)
 # define MZ_OS_FD_SET(n, p) MZ_FD_SET(n, p)
 # define MZ_OS_FD_CLR(n, p) MZ_FD_CLR(n, p)
@@ -212,8 +219,6 @@ static int udp_evt_check_ready(Scheme_Object *uw, Scheme_Schedule_Info *sinfo);
 static void udp_evt_needs_wakeup(Scheme_Object *_uw, void *fds);
 #endif
 
-static void register_tcp_listener_sync();
-
 #ifdef MZ_PRECISE_GC
 static void register_traversers(void);
 #endif
@@ -228,182 +233,42 @@ void scheme_init_network(Scheme_Env *env)
 
   netenv = scheme_primitive_module(scheme_intern_symbol("#%network"), env);
 
-  scheme_add_global_constant("tcp-connect", 
-			     scheme_make_prim_w_arity2(tcp_connect,
-						       "tcp-connect", 
-						       2, 4,
-						       2, 2), 
-			     netenv);
-  scheme_add_global_constant("tcp-connect/enable-break", 
-			     scheme_make_prim_w_arity2(tcp_connect_break,
-						       "tcp-connect/enable-break", 
-						       2, 4,
-						       2, 2), 
-			     netenv);
-  scheme_add_global_constant("tcp-listen", 
-			     scheme_make_prim_w_arity(tcp_listen,
-						      "tcp-listen", 
-						      1, 4),
-			     netenv);
-  scheme_add_global_constant("tcp-close", 
-			     scheme_make_prim_w_arity(tcp_stop,
-						      "tcp-close", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("tcp-accept-ready?", 
-			     scheme_make_prim_w_arity(tcp_accept_ready,
-						      "tcp-accept-ready?", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("tcp-accept", 
-			     scheme_make_prim_w_arity2(tcp_accept,
-						       "tcp-accept", 
-						       1, 1,
-						       2, 2), 
-			     netenv);
-  scheme_add_global_constant("tcp-accept-evt", 
-			     scheme_make_prim_w_arity(tcp_accept_evt,
-						      "tcp-accept-evt", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("tcp-accept/enable-break", 
-			     scheme_make_prim_w_arity2(tcp_accept_break,
-						       "tcp-accept/enable-break", 
-						       1, 1,
-						       2, 2), 
-			     netenv);
-  scheme_add_global_constant("tcp-listener?", 
-			     scheme_make_folding_prim(tcp_listener_p,
-						      "tcp-listener?", 
-						      1, 1, 1), 
-			     netenv);
-  scheme_add_global_constant("tcp-addresses", 
-			     scheme_make_prim_w_arity2(tcp_addresses,
-						       "tcp-addresses", 
-						       1, 2,
-						       2, 4), 
-			     netenv);
-  scheme_add_global_constant("tcp-abandon-port", 
-			     scheme_make_prim_w_arity(tcp_abandon_port,
-						      "tcp-abandon-port", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("tcp-port?", 
-			     scheme_make_folding_prim(tcp_port_p,
-						      "tcp-port?", 
-						      1, 1, 1), 
-			     netenv);
+  GLOBAL_PRIM_W_ARITY2 ( "tcp-connect"               , tcp_connect              , 2 , 4 , 2 , 2 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY2 ( "tcp-connect/enable-break"  , tcp_connect_break        , 2 , 4 , 2 , 2 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "tcp-listen"                , tcp_listen               , 1 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "tcp-close"                 , tcp_stop                 , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "tcp-accept-ready?"         , tcp_accept_ready         , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY2 ( "tcp-accept"                , tcp_accept               , 1 , 1 , 2 , 2 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "tcp-accept-evt"            , tcp_accept_evt           , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY2 ( "tcp-accept/enable-break"   , tcp_accept_break         , 1 , 1 , 2 , 2 , netenv ) ;
+  GLOBAL_FOLDING_PRIM  ( "tcp-listener?"             , tcp_listener_p           , 1 , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY2 ( "tcp-addresses"             , tcp_addresses            , 1 , 2 , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "tcp-abandon-port"          , tcp_abandon_port         , 1 , 1 , netenv ) ;
+  GLOBAL_FOLDING_PRIM  ( "tcp-port?"                 , tcp_port_p               , 1 , 1 , 1 , netenv ) ;
 
-  scheme_add_global_constant("udp-open-socket", 
-			     scheme_make_prim_w_arity(make_udp,
-						      "udp-open-socket", 
-						      0, 2), 
-			     netenv);
-  scheme_add_global_constant("udp-close", 
-			     scheme_make_prim_w_arity(udp_close,
-						      "udp-close", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("udp?", 
-			     scheme_make_folding_prim(udp_p,
-						      "udp?", 
-						      1, 1, 1), 
-			     netenv);
-  scheme_add_global_constant("udp-bound?", 
-			     scheme_make_prim_w_arity(udp_bound_p,
-						      "udp-bound?", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("udp-connected?", 
-			     scheme_make_prim_w_arity(udp_connected_p,
-						      "udp-connected?", 
-						      1, 1), 
-			     netenv);
+  GLOBAL_PRIM_W_ARITY  ( "udp-open-socket"           , make_udp                 , 0 , 2 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-close"                 , udp_close                , 1 , 1 , netenv ) ;
+  GLOBAL_FOLDING_PRIM  ( "udp?"                      , udp_p                    , 1 , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-bound?"                , udp_bound_p              , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-connected?"            , udp_connected_p          , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-bind!"                 , udp_bind                 , 3 , 3 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-connect!"              , udp_connect              , 3 , 3 , netenv ) ;
 
-  scheme_add_global_constant("udp-bind!", 
-			     scheme_make_prim_w_arity(udp_bind,
-						      "udp-bind!", 
-						      3, 3), 
-			     netenv);
-  scheme_add_global_constant("udp-connect!", 
-			     scheme_make_prim_w_arity(udp_connect,
-						      "udp-connect!", 
-						      3, 3), 
-			     netenv);
+  GLOBAL_PRIM_W_ARITY  ( "udp-send-to"               , udp_send_to              , 4 , 6 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send"                  , udp_send                 , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send-to*"              , udp_send_to_star         , 4 , 6 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send*"                 , udp_send_star            , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send-to/enable-break"  , udp_send_to_enable_break , 4 , 6 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send/enable-break"     , udp_send_enable_break    , 2 , 4 , netenv ) ;
 
-  scheme_add_global_constant("udp-send-to", 
-			     scheme_make_prim_w_arity(udp_send_to,
-						      "udp-send-to", 
-						      4, 6), 
-			     netenv);
-  scheme_add_global_constant("udp-send", 
-			     scheme_make_prim_w_arity(udp_send,
-						      "udp-send", 
-						      2, 4), 
-			     netenv);
-  scheme_add_global_constant("udp-send-to*", 
-			     scheme_make_prim_w_arity(udp_send_to_star,
-						      "udp-send-to*", 
-						      4, 6), 
-			     netenv);
-  scheme_add_global_constant("udp-send*", 
-			     scheme_make_prim_w_arity(udp_send_star,
-						      "udp-send*", 
-						      2, 4), 
-			     netenv);
-  scheme_add_global_constant("udp-send-to/enable-break", 
-			     scheme_make_prim_w_arity(udp_send_to_enable_break,
-						      "udp-send-to/enable-break", 
-						      4, 6), 
-			     netenv);
-  scheme_add_global_constant("udp-send/enable-break", 
-			     scheme_make_prim_w_arity(udp_send_enable_break,
-						      "udp-send/enable-break", 
-						      2, 4), 
-			     netenv);
-
-  scheme_add_global_constant("udp-receive!", 
-			     scheme_make_prim_w_arity(udp_receive,
-						      "udp-receive!", 
-						      2, 4), 
-			     netenv);
-  scheme_add_global_constant("udp-receive!*", 
-			     scheme_make_prim_w_arity(udp_receive_star,
-						      "udp-receive!*", 
-						      2, 4), 
-			     netenv);
-  scheme_add_global_constant("udp-receive!/enable-break", 
-			     scheme_make_prim_w_arity(udp_receive_enable_break,
-						      "udp-receive!/enable-break", 
-						      2, 4), 
-			     netenv);
-  scheme_add_global_constant("udp-receive-ready-evt", 
-			     scheme_make_prim_w_arity(udp_read_ready_evt,
-						      "udp-receive-ready-evt", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("udp-send-ready-evt", 
-			     scheme_make_prim_w_arity(udp_write_ready_evt,
-						      "udp-send-ready-evt", 
-						      1, 1), 
-			     netenv);
-  scheme_add_global_constant("udp-receive!-evt", 
-			     scheme_make_prim_w_arity(udp_read_evt,
-						      "udp-receive!-evt", 
-						      2, 4), 
-			     netenv);
-  scheme_add_global_constant("udp-send-evt", 
-			     scheme_make_prim_w_arity(udp_write_evt,
-						      "udp-send-evt", 
-						      2, 4), 
-			     netenv);
-  scheme_add_global_constant("udp-send-to-evt", 
-			     scheme_make_prim_w_arity(udp_write_to_evt,
-						      "udp-send-to-evt", 
-						      4, 6), 
-			     netenv);
-
-  register_tcp_listener_sync();
+  GLOBAL_PRIM_W_ARITY  ( "udp-receive!"              , udp_receive              , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-receive!*"             , udp_receive_star         , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-receive!/enable-break" , udp_receive_enable_break , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-receive-ready-evt"     , udp_read_ready_evt       , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send-ready-evt"        , udp_write_ready_evt      , 1 , 1 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-receive!-evt"          , udp_read_evt             , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send-evt"              , udp_write_evt            , 2 , 4 , netenv ) ;
+  GLOBAL_PRIM_W_ARITY  ( "udp-send-to-evt"           , udp_write_to_evt         , 4 , 6 , netenv ) ;
 
   scheme_finish_primitive_module(netenv);
 }
@@ -417,6 +282,8 @@ void scheme_init_network(Scheme_Env *env)
 /* These two need o be outside of USE_TCP */
 #define PORT_ID_TYPE "exact integer in [1, 65535]"
 #define CHECK_PORT_ID(obj) (SCHEME_INTP(obj) && (SCHEME_INT_VAL(obj) >= 1) && (SCHEME_INT_VAL(obj) <= 65535))
+#define LISTEN_PORT_ID_TYPE "exact integer in [0, 65535]"
+#define CHECK_LISTEN_PORT_ID(obj) (SCHEME_INTP(obj) && (SCHEME_INT_VAL(obj) >= 0) && (SCHEME_INT_VAL(obj) <= 65535))
 
 #ifdef USE_TCP
 
@@ -449,7 +316,7 @@ typedef struct SOCKADDR_IN mz_unspec_address;
 # ifdef PROTOENT_IS_INT
 #  define PROTO_P_PROTO PROTOENT_IS_INT
 # else
-static struct protoent *proto;
+SHARED_OK static struct protoent *proto;
 #  define PROTO_P_PROTO (proto ? proto->p_proto : 0)
 # endif
 
@@ -471,11 +338,9 @@ static struct protoent *proto;
 # define mz_gai_strerror gai_strerror
 #else
 # define mzAI_PASSIVE 0
-# ifdef MZ_XFORM
-START_XFORM_SKIP;
-# endif
 static int mz_getaddrinfo(const char *nodename, const char *servname,
 			  const struct mz_addrinfo *hints, struct mz_addrinfo **res)
+  XFORM_SKIP_PROC
 {
   struct hostent *h;
 
@@ -519,17 +384,16 @@ static int mz_getaddrinfo(const char *nodename, const char *servname,
   return h_errno;
 }
 void mz_freeaddrinfo(struct mz_addrinfo *ai)
+  XFORM_SKIP_PROC
 {
   free(ai->ai_addr);
   free(ai);
 }
 const char *mz_gai_strerror(int ecode)
+  XFORM_SKIP_PROC
 {
   return hstrerror(ecode);
 }
-# ifdef MZ_XFORM
-END_XFORM_SKIP;
-# endif
 #endif
 
 #if defined(USE_WINSOCK_TCP) || defined(PTHREADS_OK_FOR_GHBN)
@@ -574,11 +438,8 @@ HANDLE ready_sema;
 int ready_fd;
 # endif
 
-#ifdef MZ_XFORM
-START_XFORM_SKIP;
-#endif
-
 static long getaddrinfo_in_thread(void *data)
+  XFORM_SKIP_PROC
 {
   int ok;
   struct mz_addrinfo *res, hints;
@@ -620,10 +481,6 @@ static long getaddrinfo_in_thread(void *data)
   return 1;
 }
 
-#ifdef MZ_XFORM
-END_XFORM_SKIP;
-#endif
-
 static void release_ghbn_lock(GHBN_Rec *rec)
 {
   ghbn_lock = 0;
@@ -794,8 +651,9 @@ struct mz_addrinfo *scheme_get_host_address(const char *address, int id, int *er
   char buf[32], *service;
   int ok;
   GC_CAN_IGNORE struct mz_addrinfo *r, hints;
+  r = NULL;
 
-  if (id) {
+  if (id >= 0) {
     service = buf;
     sprintf(buf, "%d", id);
   } else
@@ -957,8 +815,8 @@ static int tcp_check_accept(Scheme_Object *_listener)
   struct timeval time = {0, 0};
   int sr, i;
 
-  INIT_DECL_OS_FDSET(readfds);
-  INIT_DECL_OS_FDSET(exnfds);
+  INIT_DECL_OS_RD_FDSET(readfds);
+  INIT_DECL_OS_ER_FDSET(exnfds);
 
   if (LISTENER_WAS_CLOSED(listener))
     return 1;
@@ -1021,8 +879,8 @@ static int tcp_check_connect(Scheme_Object *connector_p)
   struct timeval time = {0, 0};
   int sr;
 
-  INIT_DECL_OS_FDSET(writefds);
-  INIT_DECL_OS_FDSET(exnfds);
+  INIT_DECL_OS_WR_FDSET(writefds);
+  INIT_DECL_OS_ER_FDSET(exnfds);
 
   s = *(tcp_t *)connector_p;
 
@@ -1076,8 +934,8 @@ static int tcp_check_write(Scheme_Object *port)
     struct timeval time = {0, 0};
     int sr;
     
-    INIT_DECL_OS_FDSET(writefds);
-    INIT_DECL_OS_FDSET(exnfds);
+    INIT_DECL_OS_WR_FDSET(writefds);
+    INIT_DECL_OS_ER_FDSET(exnfds);
     
     s = data->tcp;
     
@@ -1175,8 +1033,8 @@ static int tcp_byte_ready (Scheme_Input_Port *port)
   DECL_OS_FDSET(exfds);
   struct timeval time = {0, 0};
 
-  INIT_DECL_OS_FDSET(readfds);
-  INIT_DECL_OS_FDSET(exfds);
+  INIT_DECL_OS_RD_FDSET(readfds);
+  INIT_DECL_OS_ER_FDSET(exfds);
 #endif
 
   if (port->closed)
@@ -1358,7 +1216,6 @@ static void tcp_close_input(Scheme_Input_Port *port)
   closesocket(data->tcp);
 #endif
 
-  --scheme_file_open_count;
 }
 
 static int
@@ -1557,7 +1414,6 @@ static void tcp_close_output(Scheme_Output_Port *port)
   closesocket(data->tcp);
 #endif
 
-  --scheme_file_open_count;
 }
 
 static int
@@ -1646,7 +1502,6 @@ static void closesocket_w_decrement(Close_Socket_Data *csd)
   if (csd->src_addr)
     mz_freeaddrinfo(csd->src_addr);
   mz_freeaddrinfo(csd->dest_addr);  
-  --scheme_file_open_count;
 }
 #endif
 
@@ -1763,7 +1618,6 @@ static Scheme_Object *tcp_connect(int argc, Scheme_Object *argv[])
 	    errno = status;
 #endif
 
-	    scheme_file_open_count++;
 	  
 	    if (inprogress) {
 	      tcp_t *sptr;
@@ -1823,7 +1677,6 @@ static Scheme_Object *tcp_connect(int argc, Scheme_Object *argv[])
 	    } else {
 	      errid = errno;
 	      closesocket(s);
-	      --scheme_file_open_count;
 	      errpart = 6;
 	    }
 	  } else {
@@ -1869,6 +1722,27 @@ tcp_connect_break(int argc, Scheme_Object *argv[])
   return scheme_call_enable_break(tcp_connect, argc, argv);
 }
 
+
+static unsigned short get_no_portno(tcp_t socket, int *_errid)
+{
+  char here[MZ_SOCK_NAME_MAX_LEN];
+  unsigned int l = sizeof(here);
+  unsigned short no_port;
+
+  if (getsockname(socket, (struct sockaddr *)here, &l)) {
+    int errid;
+    errid = SOCK_ERRNO();
+    *_errid = errid;
+    return 0;
+  }
+
+  /* don't use ntohs, since the result is put back into another sin_port: */
+  no_port = ((struct sockaddr_in *)here)->sin_port;
+  if (!no_port)
+    *_errid = 0;
+  return no_port;
+}
+
 static Scheme_Object *
 tcp_listen(int argc, Scheme_Object *argv[])
 {
@@ -1880,8 +1754,8 @@ tcp_listen(int argc, Scheme_Object *argv[])
 #endif
   const char *address;
   
-  if (!CHECK_PORT_ID(argv[0]))
-    scheme_wrong_type("tcp-listen", PORT_ID_TYPE, 0, argc, argv);
+  if (!CHECK_LISTEN_PORT_ID(argv[0]))
+    scheme_wrong_type("tcp-listen", LISTEN_PORT_ID_TYPE, 0, argc, argv);
   if (argc > 1) {
     if (!SCHEME_INTP(argv[1]) || (SCHEME_INT_VAL(argv[1]) < 1))
       scheme_wrong_type("tcp-listen", "small positive integer", 1, argc, argv);
@@ -1953,6 +1827,9 @@ tcp_listen(int argc, Scheme_Object *argv[])
 	 IPv6 doesn't work right. */
       int v6_loop = (any_v6 && any_v4), skip_v6 = 0;
 #endif
+      int first_time = 1;
+      int first_was_zero = 0;
+      unsigned short no_port = 0;
 
       errid = 0;
       for (addr = tcp_listen_addr; addr; ) {
@@ -2021,7 +1898,22 @@ tcp_listen(int argc, Scheme_Object *argv[])
 	    setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)(&reuse), sizeof(int));
 	  }
       
+          if (first_was_zero) {
+            ((struct sockaddr_in *)addr->ai_addr)->sin_port = no_port;
+          }
 	  if (!bind(s, addr->ai_addr, addr->ai_addrlen)) {
+            if (first_time) {
+              if (((struct sockaddr_in *)addr->ai_addr)->sin_port == 0) {
+                no_port = get_no_portno(s, &errid);
+                first_was_zero = 1;
+		if (no_port == 0) {
+		  closesocket(s);
+		  break;
+		}
+              }
+              first_time = 0;
+            }
+
 	    if (!listen(s, backlog)) {
 	      if (!pos) {
 		l = scheme_malloc_tagged(sizeof(listener_t) + ((count - 1) * sizeof(tcp_t)));
@@ -2039,7 +1931,6 @@ tcp_listen(int argc, Scheme_Object *argv[])
 	      }
 	      l->s[pos++] = s;
 	    
-	      scheme_file_open_count++;
 	      REGISTER_SOCKET(s);
 
 	      if (pos == count) {
@@ -2077,7 +1968,6 @@ tcp_listen(int argc, Scheme_Object *argv[])
 	s = l->s[i];
 	UNREGISTER_SOCKET(s);
 	closesocket(s);
-	--scheme_file_open_count;
       }
       
       mz_freeaddrinfo(tcp_listen_addr);
@@ -2120,7 +2010,6 @@ static int stop_listener(Scheme_Object *o)
 	UNREGISTER_SOCKET(s);
 	closesocket(s);
 	listener->s[i] = INVALID_SOCKET;
-	--scheme_file_open_count;
       }
       scheme_remove_managed(((listener_t *)o)->mref, o);
     }
@@ -2258,7 +2147,6 @@ do_tcp_accept(int argc, Scheme_Object *argv[], Scheme_Object *cust, char **_fail
     v[0] = make_tcp_input_port(tcp, "tcp-accepted", cust);
     v[1] = make_tcp_output_port(tcp, "tcp-accepted", cust);
 
-    scheme_file_open_count++;
     REGISTER_SOCKET(s);
     
     return scheme_values(2, v);
@@ -2290,7 +2178,7 @@ tcp_accept_break(int argc, Scheme_Object *argv[])
   return scheme_call_enable_break(tcp_accept, argc, argv);
 }
 
-static void register_tcp_listener_sync()
+void register_network_evts()
 {
 #ifdef USE_TCP
   scheme_add_evt(scheme_listener_type, tcp_check_accept, tcp_accept_needs_wakeup, NULL, 0);
@@ -2338,13 +2226,19 @@ static int extract_svc_value(char *svc_buf)
   return id;
 }
 
+#define SCHEME_LISTEN_PORTP(p) SAME_TYPE(SCHEME_TYPE(p), scheme_listener_type)
+#define SCHEME_UDP_PORTP(p) SAME_TYPE(SCHEME_TYPE(p), scheme_udp_type)
+
 static Scheme_Object *tcp_addresses(int argc, Scheme_Object *argv[])
 {
 #ifdef USE_TCP
+  tcp_t socket = 0;
   Scheme_Tcp *tcp = NULL;
   int closed = 0;
   Scheme_Object *result[4];
   int with_ports = 0;
+  int listener = 0;
+  int udp = 0;
 
   if (SCHEME_OUTPUT_PORTP(argv[0])) {
     Scheme_Output_Port *op;
@@ -2363,8 +2257,20 @@ static Scheme_Object *tcp_addresses(int argc, Scheme_Object *argv[])
   if (argc > 1)
     with_ports = SCHEME_TRUEP(argv[1]);
 
-  if (!tcp)
-    scheme_wrong_type("tcp-addresses", "tcp-port", 0, argc, argv);
+  if (tcp) {
+    socket = tcp->tcp;
+  }
+  else {
+    if (SCHEME_LISTEN_PORTP(argv[0])) {
+      listener = 1;
+      socket = ((listener_t *)argv[0])->s[0];
+    } else if (SCHEME_UDP_PORTP(argv[0])) {
+      udp = 1;
+      socket = ((Scheme_UDP *)argv[0])->s;
+    } else {
+      scheme_wrong_type("tcp-addresses", "tcp-port", 0, argc, argv);
+    }
+  }
 
   if (closed)
     scheme_raise_exn(MZEXN_FAIL_NETWORK,
@@ -2376,23 +2282,26 @@ static Scheme_Object *tcp_addresses(int argc, Scheme_Object *argv[])
     char here[MZ_SOCK_NAME_MAX_LEN], there[MZ_SOCK_NAME_MAX_LEN];
     char host_buf[MZ_SOCK_HOST_NAME_MAX_LEN];
     char svc_buf[MZ_SOCK_SVC_NAME_MAX_LEN];
-    unsigned int here_len, there_len;
+    unsigned int here_len;
+    unsigned int there_len = 0;
+    int peerrc = 0;
 
     l = sizeof(here);
-    if (getsockname(tcp->tcp, (struct sockaddr *)here, &l)) {
+    if (getsockname(socket, (struct sockaddr *)here, &l)) {
       scheme_raise_exn(MZEXN_FAIL_NETWORK,
 		       "tcp-addresses: could not get local address (%e)",
 		       SOCK_ERRNO());
     }
     here_len = l;
 
-    l = sizeof(there);
-    if (getpeername(tcp->tcp, (struct sockaddr *)there, &l)) {
-      scheme_raise_exn(MZEXN_FAIL_NETWORK,
-		       "tcp-addresses: could not get peer address (%e)",
-		       SOCK_ERRNO());
+    if (!listener) {
+      l = sizeof(there);
+      peerrc = getpeername(socket, (struct sockaddr *)there, &l);
+      if (peerrc && !udp) {
+        scheme_raise_exn(MZEXN_FAIL_NETWORK, "tcp-addresses: could not get peer address (%e)", SOCK_ERRNO());
+      }
+      there_len = l;
     }
-    there_len = l;
 
     scheme_getnameinfo((struct sockaddr *)here, here_len, 
 		       host_buf, sizeof(host_buf),
@@ -2404,14 +2313,20 @@ static Scheme_Object *tcp_addresses(int argc, Scheme_Object *argv[])
       result[1] = scheme_make_integer(l);
     }
 
-    scheme_getnameinfo((struct sockaddr *)there, there_len, 
-		       host_buf, sizeof(host_buf),
-                       (with_ports ? svc_buf : NULL), 
-                       (with_ports ? sizeof(svc_buf) : 0));
-    result[with_ports ? 2 : 1] = scheme_make_utf8_string(host_buf);
-    if (with_ports) {
-      l = extract_svc_value(svc_buf);
-      result[3] = scheme_make_integer(l);
+    if (listener || (udp && peerrc)) {
+      result[with_ports ? 2 : 1] = scheme_make_utf8_string("0.0.0.0");
+      result[3] = scheme_make_integer(0);
+    }
+    else {
+      scheme_getnameinfo((struct sockaddr *)there, there_len, 
+          host_buf, sizeof(host_buf),
+          (with_ports ? svc_buf : NULL), 
+          (with_ports ? sizeof(svc_buf) : 0));
+      result[with_ports ? 2 : 1] = scheme_make_utf8_string(host_buf);
+      if (with_ports) {
+        l = extract_svc_value(svc_buf);
+        result[3] = scheme_make_integer(l);
+      }
     }
   }
 # else
@@ -2587,7 +2502,6 @@ void scheme_socket_to_ports(long s, const char *name, int takeover,
   *_outp = v;
   
   if (takeover) {
-    scheme_file_open_count++;
     REGISTER_SOCKET(s);
   }
 }
@@ -2785,138 +2699,125 @@ udp_connected_p(int argc, Scheme_Object *argv[])
 
 static Scheme_Object *udp_bind_or_connect(const char *name, int argc, Scheme_Object *argv[], int do_bind)
 {
-#ifdef UDP_IS_SUPPORTED
-  Scheme_UDP *udp;
-  char *address = "";
-  unsigned short origid, id;
-  GC_CAN_IGNORE struct mz_addrinfo *udp_bind_addr;
-  int errid, err;
-
-  udp = (Scheme_UDP *)argv[0];
-#endif
-
   if (!SCHEME_UDPP(argv[0]))
     scheme_wrong_type(name, "udp socket", 0, argc, argv);
 
 #ifdef UDP_IS_SUPPORTED
-  if (!SCHEME_FALSEP(argv[1]) && !SCHEME_CHAR_STRINGP(argv[1]))
-    scheme_wrong_type(name, "string or #f", 1, argc, argv);
-  if ((do_bind || !SCHEME_FALSEP(argv[2])) && !CHECK_PORT_ID(argv[2]))
-    scheme_wrong_type(name, (do_bind ? PORT_ID_TYPE : PORT_ID_TYPE " or #f"), 2, argc, argv);
-		      
-  if (SCHEME_TRUEP(argv[1])) {
-    Scheme_Object *bs;
-    bs = scheme_char_string_to_byte_string(argv[1]);
-    address = SCHEME_BYTE_STR_VAL(bs);
-  } else
-    address = NULL;
-  if (SCHEME_TRUEP(argv[2]))
-    origid = (unsigned short)SCHEME_INT_VAL(argv[2]);
-  else
-    origid = 0;
+  {
+    Scheme_UDP *udp;
+    char *address = NULL;
+    unsigned short port = 0;
+    GC_CAN_IGNORE struct mz_addrinfo *udp_bind_addr = NULL;
 
-  if (!do_bind && (SCHEME_TRUEP(argv[1]) != SCHEME_TRUEP(argv[2]))) {
-    scheme_raise_exn(MZEXN_FAIL_CONTRACT,
-		     "%s: last two arguments must be both #f or both non-#f, given: %V %V",
-		     name, argv[1], argv[2]);
-  }
+    udp = (Scheme_UDP *)argv[0];
 
-  scheme_security_check_network(name, address, origid, !do_bind);
+    if (!SCHEME_FALSEP(argv[1]) && !SCHEME_CHAR_STRINGP(argv[1]))
+      scheme_wrong_type(name, "string or #f", 1, argc, argv);
+    if (do_bind && !CHECK_LISTEN_PORT_ID(argv[2]))
+      scheme_wrong_type(name, LISTEN_PORT_ID_TYPE, 2, argc, argv);
+    if (!do_bind && !SCHEME_FALSEP(argv[2]) && !CHECK_PORT_ID(argv[2]))
+      scheme_wrong_type(name, PORT_ID_TYPE " or #f", 2, argc, argv);
 
-  if (udp->s == INVALID_SOCKET) {
-    scheme_raise_exn(MZEXN_FAIL_NETWORK,
-		     "%s: udp socket was already closed: %V",
-		     name,
-		     udp);
-    return NULL;
-  }
+    if (SCHEME_TRUEP(argv[1])) {
+      Scheme_Object *bs;
+      bs = scheme_char_string_to_byte_string(argv[1]);
+      address = SCHEME_BYTE_STR_VAL(bs);
+    }
+    if (SCHEME_TRUEP(argv[2]))
+      port = (unsigned short)SCHEME_INT_VAL(argv[2]);
 
+    if (!do_bind && (SCHEME_TRUEP(argv[1]) != SCHEME_TRUEP(argv[2]))) {
+      scheme_raise_exn(MZEXN_FAIL_CONTRACT,
+          "%s: last two arguments must be both #f or both non-#f, given: %V %V",
+          name, argv[1], argv[2]);
+    }
 
-  if (do_bind && udp->bound) {
-    scheme_raise_exn(MZEXN_FAIL_NETWORK,
-		     "%s: udp socket is already bound: %V",
-		     name,
-		     udp);
-    return NULL;
-  }
+    scheme_security_check_network(name, address, port, !do_bind);
 
-  id = origid;
-
-  if (address || id)
-    udp_bind_addr = scheme_get_host_address(address, id, &err, -1, do_bind, 0);
-  else
-    udp_bind_addr = NULL;
-
-  if (udp_bind_addr || !origid) {
-    if (do_bind) {
-      if (!bind(udp->s, udp_bind_addr->ai_addr, udp_bind_addr->ai_addrlen)) {
-	udp->bound = 1;
-	mz_freeaddrinfo(udp_bind_addr);
-	return scheme_void;
-      }
-      errid = SOCK_ERRNO();
-    } else {
-      int ok = 1;
+    if (udp->s == INVALID_SOCKET) {
+      scheme_raise_exn(MZEXN_FAIL_NETWORK, "%s: udp socket was already closed: %V", name, udp); 
+      return NULL;
+    }
+    if (do_bind && udp->bound) { 
+      scheme_raise_exn(MZEXN_FAIL_NETWORK, "%s: udp socket is already bound: %V", name, udp);
+      return NULL;
+    }
 
+    /* DISCONNECT */
+    if (SCHEME_FALSEP(argv[1]) && SCHEME_FALSEP(argv[2])) {
+      int errid = 0;
+      if (udp->connected) {
+        int ok;
 #ifdef USE_NULL_TO_DISCONNECT_UDP
-      if (!origid) {
-	if (udp->connected)
-	  ok = !connect(udp->s, NULL, 0);
-      } else
+        ok = !connect(udp->s, NULL, 0);
+#else  //#ifndef USE_NULL_TO_DISCONNECT_UDP
+        GC_CAN_IGNORE mz_unspec_address ua;
+        ua.sin_family = AF_UNSPEC;
+        ua.sin_port = 0;
+        memset(&(ua.sin_addr), 0, sizeof(ua.sin_addr));
+        memset(&(ua.sin_zero), 0, sizeof(ua.sin_zero));
+        ok = !connect(udp->s, (struct sockaddr *)&ua, sizeof(ua));
 #endif
-	{
-	  if (udp_bind_addr)
-	    ok = !connect(udp->s, udp_bind_addr->ai_addr, udp_bind_addr->ai_addrlen);
-#ifndef USE_NULL_TO_DISCONNECT_UDP
-	  else {
-	    GC_CAN_IGNORE mz_unspec_address ua;
-	    ua.sin_family = AF_UNSPEC;
-	    ua.sin_port = 0;
-	    memset(&(ua.sin_addr), 0, sizeof(ua.sin_addr));
-	    memset(&(ua.sin_zero), 0, sizeof(ua.sin_zero));
-	    ok = !connect(udp->s, (struct sockaddr *)&ua, sizeof(ua));
-	  }
-#endif
-	}
-      
-      if (!ok)
-	errid = SOCK_ERRNO();
-      else
-	errid = 0;
-
-      if (!ok && OK_DISCONNECT_ERROR(errid) && !origid) {
-	/* It's ok. We were trying to disconnect */
-	ok = 1;
+        if (!ok) errid = SOCK_ERRNO();
+        if (ok || OK_DISCONNECT_ERROR(errid)) {
+          udp->connected = 0;
+          return scheme_void;
+        }
+        else {
+          scheme_raise_exn(MZEXN_FAIL_NETWORK, "%s: can't disconnect port: %d on address: %s (%E)", name, port, address ? address : "#f", errid);
+        }
       }
+      return scheme_void;
+    }
 
-      if (ok) {
-	if (origid)
-	  udp->connected = 1;
-	else
-	  udp->connected = 0;
-	if (udp_bind_addr)
-	  mz_freeaddrinfo(udp_bind_addr);
-	return scheme_void;
+    /* RESOLVE ADDRESS */
+    if (address || port) {
+      int err;
+      udp_bind_addr = scheme_get_host_address(address, port, &err, -1, do_bind, 0);
+      if (!udp_bind_addr) {
+        scheme_raise_exn(MZEXN_FAIL_NETWORK, "%s: can't resolve address: %s (%N)", name, address, 1, err);
+        return NULL;
       }
     }
 
-    if (udp_bind_addr)
+    /* CONNECT CASE */
+    if (!do_bind) {
+      int ok = !connect(udp->s, udp_bind_addr->ai_addr, udp_bind_addr->ai_addrlen);
       mz_freeaddrinfo(udp_bind_addr);
-
-    scheme_raise_exn(MZEXN_FAIL_NETWORK,
-		     "%s: can't %s to port: %d on address: %s (%E)", 
-		     name,
-		     do_bind ? "bind" : "connect",
-		     origid,
-		     address ? address : "#f",
-		     errid);
-    return NULL;
-  } else {
-    scheme_raise_exn(MZEXN_FAIL_NETWORK,
-		     "%s: can't resolve address: %s (%N)", 
-		     name,
-		     address, 1, err);
-    return NULL;
+      if (ok) {
+        udp->connected = 1;
+        return scheme_void;
+      }
+      else {
+        scheme_raise_exn(MZEXN_FAIL_NETWORK, "%s: can't connect to port: %d on address: %s (%E)", name, port, address ? address : "#f", SOCK_ERRNO());
+        return NULL;
+      }
+    }
+    /* BIND CASE */
+    else {
+      int ok;
+      if (udp_bind_addr == NULL ) {
+        GC_CAN_IGNORE mz_unspec_address ua;
+        memset(&ua, 0, sizeof(mz_unspec_address));
+        ua.sin_family = AF_UNSPEC;
+        ua.sin_port = 0;
+        memset(&(ua.sin_addr), 0, sizeof(ua.sin_addr));
+        memset(&(ua.sin_zero), 0, sizeof(ua.sin_zero));
+        ok = !bind(udp->s, (struct sockaddr *)&ua, sizeof(ua));
+      }
+      else {
+        ok = !bind(udp->s, udp_bind_addr->ai_addr, udp_bind_addr->ai_addrlen);
+        mz_freeaddrinfo(udp_bind_addr);
+      }
+      if (ok) {
+        udp->bound = 1;
+        return scheme_void;
+      }
+      else {
+        scheme_raise_exn(MZEXN_FAIL_NETWORK, "%s: can't bind to port: %d on address: %s (%E)", name, port, address ? address : "#f", SOCK_ERRNO());
+        return NULL;
+      }
+    }
   }
 #else
   return scheme_void;
@@ -2948,8 +2849,8 @@ static int udp_check_send(Scheme_Object *_udp)
     struct timeval time = {0, 0};
     int sr;
     
-    INIT_DECL_OS_FDSET(writefds);
-    INIT_DECL_OS_FDSET(exnfds);
+    INIT_DECL_OS_WR_FDSET(writefds);
+    INIT_DECL_OS_ER_FDSET(exnfds);
     
     MZ_OS_FD_ZERO(writefds);
     MZ_OS_FD_SET(udp->s, writefds);
@@ -3180,8 +3081,8 @@ static int udp_check_recv(Scheme_Object *_udp)
     struct timeval time = {0, 0};
     int sr;
     
-    INIT_DECL_OS_FDSET(readfds);
-    INIT_DECL_OS_FDSET(exnfds);
+    INIT_DECL_OS_RD_FDSET(readfds);
+    INIT_DECL_OS_ER_FDSET(exnfds);
     
     MZ_OS_FD_ZERO(readfds);
     MZ_OS_FD_SET(udp->s, readfds);
