@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2008 PLT Scheme Inc.
+  Copyright (c) 2004-2010 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -29,6 +29,7 @@
 #endif
 
 #include "schpriv.h"
+#include "mzrt.h"
 #include <string.h>
 #include <ctype.h>
 #include "schgc.h"
@@ -47,21 +48,32 @@ extern MZ_DLLIMPORT void (*GC_custom_finalize)(void);
 extern int GC_is_marked(void *);
 #endif
 
-Scheme_Hash_Table *scheme_symbol_table = NULL;
-Scheme_Hash_Table *scheme_keyword_table = NULL;
-Scheme_Hash_Table *scheme_parallel_symbol_table = NULL;
+SHARED_OK Scheme_Hash_Table *scheme_symbol_table = NULL;
+SHARED_OK Scheme_Hash_Table *scheme_keyword_table = NULL;
+SHARED_OK Scheme_Hash_Table *scheme_parallel_symbol_table = NULL;
 
-unsigned long scheme_max_found_symbol_name;
+#ifdef MZ_USE_PLACES
+SHARED_OK static mzrt_rwlock *symbol_table_lock;
+#else
+# define mzrt_rwlock_rdlock(l) /* empty */
+# define mzrt_rwlock_wrlock(l) /* empty */
+# define mzrt_rwlock_unlock(l) /* empty */
+#endif
+
+SHARED_OK static unsigned long scheme_max_symbol_length;
 
 /* globals */
-int scheme_case_sensitive = 1;
+SHARED_OK int scheme_case_sensitive = 1;
+THREAD_LOCAL_DECL(static int gensym_counter);
 
 void scheme_set_case_sensitive(int v) { scheme_case_sensitive =  v; }
 
 /* locals */
 static Scheme_Object *symbol_p_prim (int argc, Scheme_Object *argv[]);
+static Scheme_Object *symbol_interned_p_prim (int argc, Scheme_Object *argv[]);
 static Scheme_Object *string_to_symbol_prim (int argc, Scheme_Object *argv[]);
 static Scheme_Object *string_to_uninterned_symbol_prim (int argc, Scheme_Object *argv[]);
+static Scheme_Object *string_to_unreadable_symbol_prim (int argc, Scheme_Object *argv[]);
 static Scheme_Object *symbol_to_string_prim (int argc, Scheme_Object *argv[]);
 static Scheme_Object *keyword_p_prim (int argc, Scheme_Object *argv[]);
 static Scheme_Object *keyword_lt (int argc, Scheme_Object *argv[]);
@@ -69,7 +81,6 @@ static Scheme_Object *string_to_keyword_prim (int argc, Scheme_Object *argv[]);
 static Scheme_Object *keyword_to_string_prim (int argc, Scheme_Object *argv[]);
 static Scheme_Object *gensym(int argc, Scheme_Object *argv[]);
 
-static int gensym_counter;
 
 /**************************************************************************/
 
@@ -244,6 +255,9 @@ static void clean_symbol_table(void)
 # ifndef MZ_PRECISE_GC
   scheme_clean_cust_box_list();
 # endif
+# ifndef MZ_PRECISE_GC
+  scheme_notify_code_gc();
+# endif
 }
 #endif
 
@@ -282,6 +296,10 @@ scheme_init_symbol_table ()
   scheme_keyword_table = init_one_symbol_table();
   scheme_parallel_symbol_table = init_one_symbol_table();
 
+#ifdef MZ_USE_PLACES
+  mzrt_rwlock_create(&symbol_table_lock);
+#endif
+
 #ifndef MZ_PRECISE_GC
   GC_custom_finalize = clean_symbol_table;
 #endif
@@ -300,49 +318,27 @@ scheme_init_symbol (Scheme_Env *env)
   p = scheme_make_folding_prim(symbol_p_prim, "symbol?", 1, 1, 1);
   SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;
   scheme_add_global_constant("symbol?", p, env);
-
-  scheme_add_global_constant("string->symbol",
-			     scheme_make_immed_prim(string_to_symbol_prim,
-						    "string->symbol",
-						    1, 1), env);
-  scheme_add_global_constant("string->uninterned-symbol",
-			     scheme_make_immed_prim(string_to_uninterned_symbol_prim,
-						    "string->uninterned-symbol",
-						    1, 1),
-			     env);
-  scheme_add_global_constant("symbol->string",
-			     scheme_make_immed_prim(symbol_to_string_prim,
-						    "symbol->string",
-						    1, 1),
-			     env);
-
-  scheme_add_global_constant("keyword?",
-			     scheme_make_folding_prim(keyword_p_prim,
-						      "keyword?",
-						      1, 1, 1),
-			     env);
-  scheme_add_global_constant("keyword<?",
-			     scheme_make_folding_prim(keyword_lt,
-						      "keyword<?",
-						      2, -1, 1),
-			     env);
-  scheme_add_global_constant("string->keyword",
-			     scheme_make_immed_prim(string_to_keyword_prim,
-						    "string->keyword",
-						    1, 1), env);
-  scheme_add_global_constant("keyword->string",
-			     scheme_make_immed_prim(keyword_to_string_prim,
-						    "keyword->string",
-						    1, 1),
-			     env);
-
-  scheme_add_global_constant("gensym",
-			     scheme_make_immed_prim(gensym,
-						    "gensym",
-						    0, 1),
-			     env);
+  
+  p = scheme_make_folding_prim(symbol_interned_p_prim, "symbol-interned?", 1, 1, 1);
+  scheme_add_global_constant("symbol-interned?", p, env);
+  
+  GLOBAL_IMMED_PRIM("string->symbol",             string_to_symbol_prim,            1, 1, env);
+  GLOBAL_IMMED_PRIM("string->uninterned-symbol",  string_to_uninterned_symbol_prim, 1, 1, env);
+  GLOBAL_IMMED_PRIM("string->unreadable-symbol",  string_to_unreadable_symbol_prim, 1, 1, env);
+  GLOBAL_IMMED_PRIM("symbol->string",             symbol_to_string_prim,            1, 1, env);
+  GLOBAL_FOLDING_PRIM("keyword?",                 keyword_p_prim,                   1, 1, 1, env);
+  GLOBAL_FOLDING_PRIM("keyword<?",                keyword_lt,                       2, -1, 1, env);
+  GLOBAL_IMMED_PRIM("string->keyword",            string_to_keyword_prim,           1, 1, env);
+  GLOBAL_IMMED_PRIM("keyword->string",            keyword_to_string_prim,           1, 1, env);
+  GLOBAL_IMMED_PRIM("gensym",                     gensym,                           0, 1, env);
 }
 
+unsigned long scheme_get_max_symbol_length() {
+  /* x86, x86_64, and powerpc support aligned_atomic_loads_and_stores */
+  return scheme_max_symbol_length;
+}
+
+
 static Scheme_Object *
 make_a_symbol(const char *name, unsigned int len, int kind)
 {
@@ -356,10 +352,13 @@ make_a_symbol(const char *name, unsigned int len, int kind)
   memcpy(sym->s, name, len);
   sym->s[len] = 0;
 
-  if (len > scheme_max_found_symbol_name) {
-    scheme_max_found_symbol_name = len;
-    scheme_reset_prepared_error_buffer();
+#ifdef MZ_USE_PLACES
+  mzrt_ensure_max_cas(&scheme_max_symbol_length, len);
+#else
+  if ( len > scheme_max_symbol_length ) {
+    scheme_max_symbol_length = len;
   }
+#endif
 
   return (Scheme_Object *) sym;
 }
@@ -386,20 +385,45 @@ scheme_make_exact_char_symbol(const mzchar *name, unsigned int len)
 }
 
 Scheme_Object *
-scheme_intern_exact_symbol_in_table(Scheme_Hash_Table *symbol_table, int kind, const char *name, unsigned int len)
+scheme_intern_exact_symbol_in_table_worker(Scheme_Hash_Table *symbol_table, int kind, const char *name, unsigned int len)
 {
   Scheme_Object *sym;
 
+  mzrt_rwlock_rdlock(symbol_table_lock);
   sym = symbol_bucket(symbol_table, name, len, NULL);
+  mzrt_rwlock_unlock(symbol_table_lock);
 
   if (!sym) {
-    sym = make_a_symbol(name, len, kind);
-    symbol_bucket(symbol_table, name, len, sym);
+    Scheme_Object *newsymbol;
+    newsymbol = make_a_symbol(name, len, kind);
+    
+    /* we must return the result of this symbol bucket call because another
+     * thread could have inserted the same symbol between the first
+     * :qsymbol_bucket call above and this one */
+    mzrt_rwlock_wrlock(symbol_table_lock);
+    sym = symbol_bucket(symbol_table, name, len, newsymbol);
+    mzrt_rwlock_unlock(symbol_table_lock);
   }
 
   return sym;
 }
 
+Scheme_Object *
+scheme_intern_exact_symbol_in_table(Scheme_Hash_Table *symbol_table, int kind, const char *name, unsigned int len)
+{
+#if defined(MZ_USE_PLACES) && defined(MZ_PRECISE_GC)
+  void *return_payload;
+  Scheme_Symbol_Parts parts;
+  parts.table = symbol_table;
+  parts.kind = kind;
+  parts.len  = len;
+  parts.name = name;
+  return_payload = scheme_master_fast_path(3, &parts);
+  return (Scheme_Object*) return_payload;
+#endif
+  return scheme_intern_exact_symbol_in_table_worker(symbol_table, kind, name, len);
+}
+
 Scheme_Object *
 scheme_intern_exact_symbol(const char *name, unsigned int len)
 {
@@ -654,6 +678,16 @@ symbol_p_prim (int argc, Scheme_Object *argv[])
   return SCHEME_SYMBOLP(argv[0]) ? scheme_true : scheme_false;
 }
 
+static Scheme_Object *
+symbol_interned_p_prim (int argc, Scheme_Object *argv[])
+{
+  if (SCHEME_SYMBOLP(argv[0]))
+    return (SCHEME_SYM_WEIRDP(argv[0]) ? scheme_false : scheme_true);
+  
+  scheme_wrong_type("symbol-interned?", "symbol", 0, argc, argv);
+  return NULL;
+}
+
 static Scheme_Object *
 string_to_symbol_prim (int argc, Scheme_Object *argv[])
 {
@@ -672,6 +706,22 @@ string_to_uninterned_symbol_prim (int argc, Scheme_Object *argv[])
 				       SCHEME_CHAR_STRTAG_VAL(argv[0]));
 }
 
+static Scheme_Object *
+string_to_unreadable_symbol_prim (int argc, Scheme_Object *argv[])
+{
+  char buf[64], *bs;
+  long blen;
+
+  if (!SCHEME_CHAR_STRINGP(argv[0]))
+    scheme_wrong_type("string->symbol", "string", 0, argc, argv);
+
+  bs = scheme_utf8_encode_to_buffer_len(SCHEME_CHAR_STR_VAL(argv[0]),
+                                        SCHEME_CHAR_STRTAG_VAL(argv[0]), 
+                                        buf, 64, &blen);
+
+  return scheme_intern_exact_parallel_symbol(bs, blen);
+}
+
 static Scheme_Object *
 symbol_to_string_prim (int argc, Scheme_Object *argv[])
 {
